# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-28 10:23
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0021_entity_descriptor_completed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='data',
            name='checksum',
            field=models.CharField(max_length=64, validators=[django.core.validators.RegexValidator(code='invalid_checksum', message='Checksum is exactly 40 alphanumerics', regex='^[0-9a-f]{64}$')]),
        ),
    ]
