# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-15 06:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0008_fix_jsonfields'),
    ]

    operations = [
        migrations.AddField(
            model_name='data',
            name='parents',
            field=models.ManyToManyField(related_name='children', to='flow.Data'),
        ),
    ]
