# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-04-12 04:33
from __future__ import unicode_literals

import autoslug.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0003_support_sample'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='data',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='descriptorschema',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='process',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='storage',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('version',), max_length=100),
        ),
    ]
