# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-08 06:56
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0002_project_to_collection'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='descriptor',
            field=django.contrib.postgres.fields.jsonb.JSONField(default={}),
        ),
        migrations.AddField(
            model_name='collection',
            name='descriptor_schema',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='flow.DescriptorSchema'),
        ),
        migrations.AddField(
            model_name='process',
            name='flow_collection',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]
