""".. Ignore pydocstyle D400.

====
Flow
====

"""
from __future__ import absolute_import, division, print_function, unicode_literals

default_app_config = 'resolwe.flow.apps.FlowConfig'  # pylint: disable=invalid-name
