"""Workflow execution engine exceptions."""
from __future__ import absolute_import, division, print_function, unicode_literals


class ExecutionError(Exception):
    """Raised for errors that occur during execution engine evaluation."""

    pass
