"""Workload management system connectors.

The connectors are used as glue between the Resolwe Manager and various
concrete workload management systems that might be used by it. Since the
only functional requirement is job submission, they can be simple and
nearly contextless.
"""
