"""Workflow expression engine exceptions."""
from __future__ import absolute_import, division, print_function, unicode_literals


class EvaluationError(Exception):
    """Raised for errors that occur during expression evaluation."""

    pass
